
// 2/23/12 first quicl attempt to get andy's galaxy code based off of some of the CUDA then Onvidia OpenCL code
// to be happy in the AMD world by crashing it into apple's julia set openCL example
// this version tweaks the code to make it happy on CPUs and GPUs by introducing USE_GPU

#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <OpenGL/OpenGL.h>
#include <OpenGL/gl.h>
#include <OpenGL/CGLDevice.h>
#include <GLUT/glut.h>
#include <OpenCL/opencl.h>

#include <mach/mach_time.h>


static void Cleanup(void);
static void Shutdown(void);
void timerEvent(int value);
static int Initialize(int gpu);
void runKernel();

#define REFRESH_DELAY	  5 //ms

cl_int ciErrNum;

////////////////////////////////////////////////////////////////////////////////

#define USE_GL_ATTACHMENTS              (0) // enable OpenGL attachments for Compute results
#define USE_GPU                         (1) 
#define DEBUG_INFO                      (0)     
#define COMPUTE_KERNEL_FILENAME         ("oclAndyGalaxy.cl")
#define COMPUTE_KERNEL_METHOD_NAME      ("galaxy")
#define SEPARATOR                       ("----------------------------------------------------------------------\n")


////////////////////////////////////////////////////////////////////////////////

static cl_context                       ComputeContext;
static cl_command_queue                 ComputeCommands;
static cl_kernel                        ComputeKernel;
static cl_program                       ComputeProgram;
static cl_device_id                     ComputeDeviceId;
static cl_device_type                   ComputeDeviceType;


////////////////////////////////////////////////////////////////////////////////

#define WIDTH 512
#define HEIGHT 512

static int Width                        = WIDTH;
static int Height                       = HEIGHT;



////////////////////////////////////////////////////////////////////////////////

const unsigned int window_width = 512;
const unsigned int window_height = 512;

int iGLUTWindowHandle = 0;          // handle to the GLUT window
unsigned int anim = 0;

// vbo variables
GLuint vbo;

// mouse controls
int mouse_old_x, mouse_old_y;
int mouse_buttons = 0;
float rotate_x = 0.0, rotate_y = 0.0;

// galaxy variables
float * d_particleData;
float * h_particleData;

float targetX, targetY, targetZ;

float origStars[82000][7];
float stars[82000][7];

int origNumStars = 0;
int numStars;

int mode = 3; // how many stars to skip - was 5
int nd = 7;

int computationHeight = 1;

cl_mem vbo_cl;
cl_mem cl_membufferData;

//*****************************************************************************

static void Display(void)
{
    anim += 1;
    
    // run OpenCL kernel to generate vertex positions
    runKernel();
    
    // set view matrix
    glClear(GL_COLOR_BUFFER_BIT);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glRotatef(rotate_x, 1.0, 0.0, 0.0);
    glRotatef(rotate_y, 0.0, 1.0, 0.0);
    
    
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glVertexPointer(4, GL_FLOAT, 0, 0);
    glColorPointer(4, GL_FLOAT, 0, (GLvoid *) (numStars * sizeof(float)*4));
    
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    
    
    glDrawArrays(GL_POINTS, 0, numStars);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    
    // flip backbuffer to screen
    glutSwapBuffers();
    glutPostRedisplay();
    
}

//*****************************************************************************

static void  Reshape (int w, int h)
{
    glViewport(0, 0, w, h);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glClear(GL_COLOR_BUFFER_BIT);
    glutSwapBuffers();
    
    Width = w;
    Height = h;    
}

void Keyboard( unsigned char key, int x, int y )
{
    
    switch( key )
    {
        case 27:
            exit(0);
            break;
            
    }
    
}

void timerEvent(int value)
{
    glutPostRedisplay();
	glutTimerFunc(REFRESH_DELAY, timerEvent,0);
}


//*****************************************************************************

int loadInStars(void)
{
	FILE * starFile = NULL;
	char line[255];
    
	float mass, x, y, z, vx, vy, vz;
    
	starFile = fopen("galaxy.txt", "r");
    
	if (starFile == NULL)
    {
		fprintf(stderr, "can not find galaxy data file\n");
		return(0);
    }
    
	numStars = 0;
	origNumStars = 0;
    
	fgets(line, 255, starFile);
    
	while (!feof(starFile))
    {
		fgets(line, 255, starFile);
        
		mass    = 0.0f;
		sscanf(line, "%f %f %f %f %f %f %f",
               &mass, &x, &y, &z, &vx, &vy, &vz);
        
		origStars[origNumStars][0] = x;
		origStars[origNumStars][1] = y;
		origStars[origNumStars][2] = z;
		origStars[origNumStars][3] = vx;
		origStars[origNumStars][4] = vy;
		origStars[origNumStars][5] = vz;
		origStars[origNumStars][6] = mass;
        
		origNumStars++;
    }
    
	fclose(starFile);
	fprintf(stderr, "%d stars found\n", origNumStars);
    
	int s;
    
	for (s=0; s< 49152; s+= (1+mode)) // exclude the halo stars
	{
        stars[numStars][0] = origStars[s][0];
        stars[numStars][1] = origStars[s][1];
        stars[numStars][2] = origStars[s][2];
        stars[numStars][3] = origStars[s][3];
        stars[numStars][4] = origStars[s][4];
        stars[numStars][5] = origStars[s][5];
        stars[numStars][6] = origStars[s][6];
        
        numStars += 1;
	}    
    
	fprintf(stderr, "viewing %d stars\n", numStars);
    
    // 81920 total stars in file
    // 49152 stars excluding the halo stars
    // max stars might be around 32000 on this machine
    
	return(1);
}


//*****************************************************************************

static int LoadTextFromFile(
    const char *file_name, char **result_string, size_t *string_len)
{
    int fd;
    unsigned file_len;
    struct stat file_status;
    int ret;

    *string_len = 0;
    fd = open(file_name, O_RDONLY);
    if (fd == -1)
    {
        printf("Error opening file %s\n", file_name);
        return -1;
    }
    ret = fstat(fd, &file_status);
    if (ret)
    {
        printf("Error reading status for file %s\n", file_name);
        return -1;
    }
    file_len = file_status.st_size;

    *result_string = (char*)calloc(file_len + 1, sizeof(char));
    ret = read(fd, *result_string, file_len);
    if (!ret)
    {
        printf("Error reading from file %s\n", file_name);
        return -1;
    }

    close(fd);

    *string_len = file_len;
    return 0;
}

//*****************************************************************************


static int  SetupComputeDevices(int gpu)
{
    int err;
	size_t returned_size;
    ComputeDeviceType = gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU;

#if (USE_GL_ATTACHMENTS && USE_GPU)

    printf(SEPARATOR);
    printf("Using active OpenGL context...\n");

    CGLContextObj myContext = CGLGetCurrentContext();              
    CGLShareGroupObj myShareGroup = CGLGetShareGroup(myContext);
    
    cl_context_properties properties[] = { 
        CL_CONTEXT_PROPERTY_USE_CGL_SHAREGROUP_APPLE, 
        (cl_context_properties)myShareGroup, 0 
    };
        
    // Create a context from a CGL share group
    //
    ComputeContext = clCreateContext(properties, 0, 0, clLogMessagesToStdoutAPPLE, 0, 0);
    if (!ComputeContext)
    {
        printf("Error: Failed to create a compute context!\n");
        return EXIT_FAILURE;
    }

#else

    // Locate a compute device
    //
    err = clGetDeviceIDs(NULL, ComputeDeviceType, 1, &ComputeDeviceId, NULL);
    if (err != CL_SUCCESS)
    {
        printf("Error: Failed to locate compute device!\n");
        return EXIT_FAILURE;
    }
  
    // Create a context containing the compute device(s)
    //
    ComputeContext = clCreateContext(0, 1, &ComputeDeviceId, clLogMessagesToStdoutAPPLE, NULL, &err);
    if (!ComputeContext)
    {
        printf("Error: Failed to create a compute context!\n");
        return EXIT_FAILURE;
    }

#endif

    unsigned int device_count;
    cl_device_id device_ids[16];

    err = clGetContextInfo(ComputeContext, CL_CONTEXT_DEVICES, sizeof(device_ids), device_ids, &returned_size);
    if(err)
    {
        printf("Error: Failed to retrieve compute devices for context!\n");
        return EXIT_FAILURE;
    }
    
    device_count = returned_size / sizeof(cl_device_id);
    
    int i = 0;
    int device_found = 0;
    cl_device_type device_type;
    for(i = 0; i < device_count; i++) 
    {
        clGetDeviceInfo(device_ids[i], CL_DEVICE_TYPE, sizeof(cl_device_type), &device_type, NULL);
        if(device_type == ComputeDeviceType) 
        {
            ComputeDeviceId = device_ids[i];
            device_found = 1;
            break;
        }	
    }
    
    if(!device_found) // CPU version failing here
    {
        printf("Error: Failed to locate compute device!\n");
        return EXIT_FAILURE;
    }
        
    // Create a command queue
    //
    ComputeCommands = clCreateCommandQueue(ComputeContext, ComputeDeviceId, CL_QUEUE_PROFILING_ENABLE, &err);
    if (!ComputeCommands)
    {
        printf("Error: Failed to create a command queue!\n");
        return EXIT_FAILURE;
    }

    // Report the device vendor and device name
    // 
    cl_char vendor_name[1024] = {0};
    cl_char device_name[1024] = {0};
    err = clGetDeviceInfo(ComputeDeviceId, CL_DEVICE_VENDOR, sizeof(vendor_name), vendor_name, &returned_size);
    err|= clGetDeviceInfo(ComputeDeviceId, CL_DEVICE_NAME, sizeof(device_name), device_name, &returned_size);
    if (err != CL_SUCCESS)
    {
        printf("Error: Failed to retrieve device info!\n");
        return EXIT_FAILURE;
    }

    printf(SEPARATOR);
    printf("Connecting to %s %s...\n", vendor_name, device_name);

    return CL_SUCCESS;
}

//*****************************************************************************

static int  SetupComputeKernel(void)
{
    int err = 0;
    char *source = 0;
    size_t length = 0;

    if(ComputeKernel)
        clReleaseKernel(ComputeKernel);    
    ComputeKernel = 0;

    if(ComputeProgram)
        clReleaseProgram(ComputeProgram);
    ComputeProgram = 0;
    
    printf(SEPARATOR);
    printf("Loading kernel source from file '%s'...\n", COMPUTE_KERNEL_FILENAME);    
    err = LoadTextFromFile(COMPUTE_KERNEL_FILENAME, &source, &length);
    if (!source || err)
    {
        printf("Error: Failed to load kernel source!\n");
        return EXIT_FAILURE;
    }

    const char* width_macro = "#define WIDTH";
    const char* height_macro = "#define HEIGHT";
    
    char* preprocess = malloc(strlen(source) + 1024);
    sprintf(preprocess, "\n%s (%d)\n%s (%d)\n%s", width_macro, Width, height_macro, Height, source); 
#if (DEBUG_INFO)
    printf("%s", preprocess);
#endif

    // Create the compute program from the source buffer
    //
    ComputeProgram = clCreateProgramWithSource(ComputeContext, 1, (const char **) & preprocess, NULL, &err);
    if (!ComputeProgram || err != CL_SUCCESS)
    {
        printf("Error: Failed to create compute program!\n");
        return EXIT_FAILURE;
    }
    free(source);
    free(preprocess);

    // Build the program executable - THIS SEEMS TO FAIL FOR SOME -
    //
    err = clBuildProgram(ComputeProgram, 0, NULL, NULL, NULL, NULL);
    if (err != CL_SUCCESS)
    {
        size_t len;
        char buffer[2048];

        printf("Error: Failed to build program executable!\n");
        clGetProgramBuildInfo(ComputeProgram, ComputeDeviceId, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
        printf("%s\n", buffer);
        return EXIT_FAILURE;
    }

    // Create the compute kernel from within the program
    //
    printf("Creating kernel '%s'...\n", COMPUTE_KERNEL_METHOD_NAME);    
    ComputeKernel = clCreateKernel(ComputeProgram, COMPUTE_KERNEL_METHOD_NAME, &err);
    if (!ComputeKernel || err != CL_SUCCESS)
    {
        printf("Error: Failed to create compute kernel!\n");
        return EXIT_FAILURE;
    }

    return CL_SUCCESS;

}

//*****************************************************************************

static void Cleanup(void)
{
    clFinish(ComputeCommands);
    clReleaseKernel(ComputeKernel);
    clReleaseProgram(ComputeProgram);
    clReleaseCommandQueue(ComputeCommands);
    clReleaseContext(ComputeContext);
    
    ComputeCommands = 0;
    ComputeKernel = 0;
    ComputeProgram = 0;    
    ComputeContext = 0;
}

static void
Shutdown(void)
{
    printf(SEPARATOR);
    printf("Shutting down...\n");
    Cleanup();
    exit(0);
}

// Mouse event handlers
//*****************************************************************************
void mouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN) {
        mouse_buttons |= 1<<button;
    } else if (state == GLUT_UP) {
        mouse_buttons = 0;
    }
    
    mouse_old_x = x;
    mouse_old_y = y;
}

void motion(int x, int y)
{
    float dx, dy;
    dx = (float)(x - mouse_old_x);
    dy = (float)(y - mouse_old_y);
    
    if (mouse_buttons & 1) {
        rotate_x += dy * 0.2f;
        rotate_y += dx * 0.2f;
    } 
    
    mouse_old_x = x;
    mouse_old_y = y;
    
}


void Idle(void)
{
    glutPostRedisplay();
}

////////////////////////////////////////////////////////////////////////////////

static int SetupGraphics(void)
{
 
    // initialize GLUT 
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutInitWindowPosition (glutGet(GLUT_SCREEN_WIDTH)/2 - window_width/2, 
                            glutGet(GLUT_SCREEN_HEIGHT)/2 - window_height/2);
    glutInitWindowSize(window_width, window_height);
    iGLUTWindowHandle = glutCreateWindow("OpenCL/GL Interop (VBO)");
#if !(defined (__APPLE__) || defined(MACOSX))
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);
#endif
    
    // register GLUT callback functions
    glutDisplayFunc(Display);
    glutKeyboardFunc(Keyboard);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
	glutTimerFunc(REFRESH_DELAY, timerEvent,0);
    
    
    // default initialization
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glDisable(GL_DEPTH_TEST);
    
 	glPointSize(2.0);
    
    // viewport
    glViewport(0, 0, window_width, window_height);
    
    // projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, (GLfloat)window_width / (GLfloat) window_height, 0.1, 1000.0);
    gluLookAt(0, 0, 40,    0, 0, 0,    0, 1, 0);
    
    return(GL_NO_ERROR);
}


// Run the OpenCL part of the computation
//*****************************************************************************
void runKernel()
{
	size_t szGlobalWorkSize[] = {numStars/computationHeight, computationHeight};
    
    ciErrNum = CL_SUCCESS;
    
#ifdef GL_INTEROP   
    // map OpenGL buffer object for writing from OpenCL
    glFinish();
    ciErrNum = clEnqueueAcquireGLObjects(ComputeCommands, 1, &vbo_cl, 0,0,0);
    shrCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
#endif
    
    
    cl_event eventGlobal;
    cl_int errcode_ret;
    
    // Set arg 4 and execute the kernel
    ciErrNum = clSetKernelArg(ComputeKernel, 5, sizeof(unsigned int), &anim);
    ciErrNum |= clEnqueueNDRangeKernel(ComputeCommands, ComputeKernel, 2, NULL, szGlobalWorkSize, NULL, 0, 0, &eventGlobal );
    cl_ulong end, start;
    
    errcode_ret = clWaitForEvents(1, &eventGlobal);
    //oclCheckError(errcode_ret, CL_SUCCESS);
    errcode_ret = clGetEventProfilingInfo(eventGlobal, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, 0);
    errcode_ret |= clGetEventProfilingInfo(eventGlobal, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, 0);
    //oclCheckError(errcode_ret, CL_SUCCESS);
    fprintf(stderr, "time in kernel: %0.2f ms\n",(end-start)*1.0e-6f);

     
    
#ifdef GL_INTEROP
    // unmap buffer object
    ciErrNum = clEnqueueReleaseGLObjects(ComputeCommands, 1, &vbo_cl, 0,0,0);
    shrCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
    clFinish(ComputeCommands);
#else
    
    // Explicit Copy 
    // map the PBO to copy data from the CL buffer via host
    glBindBufferARB(GL_ARRAY_BUFFER, vbo);    
    
    // map the buffer object into client's memory
    void* ptr = glMapBufferARB(GL_ARRAY_BUFFER, GL_WRITE_ONLY_ARB);
    
    ciErrNum = clEnqueueReadBuffer(ComputeCommands, vbo_cl, CL_TRUE, 0, 8 * numStars * sizeof(float), ptr, 0, NULL, NULL);
    //shrCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
    
    glUnmapBufferARB(GL_ARRAY_BUFFER); 
#endif
    
}

//*****************************************************************************

static int Initialize(int gpu)
{
    int err;
    err = SetupGraphics();
    if (err != GL_NO_ERROR)
    {
        printf ("Failed to setup OpenGL state!");
        exit (err);
    }

    err = SetupComputeDevices(gpu);
    if(err != CL_SUCCESS)
    {
        printf ("Failed to connect to compute device! Error %d\n", err);
        exit (err);
    }
    
    err = SetupComputeKernel();
    if (err != CL_SUCCESS)
    {
        printf ("Failed to setup compute kernel! Error %d\n", err);
        exit (err);
    }
    
    return CL_SUCCESS;
}



// Create VBO
//*****************************************************************************
void createVBO(GLuint* vbo)
{
    // create VBO
    unsigned int size = numStars * 8 * sizeof( float); //4 position, 4 color
    
        // create buffer object
        glGenBuffers(1, vbo);
        glBindBuffer(GL_ARRAY_BUFFER, *vbo);
        
        // initialize buffer object
        glBufferData(GL_ARRAY_BUFFER, size, 0, GL_DYNAMIC_DRAW);
        
#ifdef GL_INTEROP
        // create OpenCL buffer from GL VBO
        vbo_cl = clCreateFromGLBuffer(ComputeContext, CL_MEM_WRITE_ONLY, *vbo, NULL);
#else
        // create standard OpenCL mem buffer
        vbo_cl = clCreateBuffer(ComputeContext, CL_MEM_WRITE_ONLY, size, NULL, &ciErrNum);
#endif
        //shrCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);
  
}

//*****************************************************************************

int main(int argc, char** argv)
{
    //read in the text file of star positions and velocities
	loadInStars();

    
    glutInit(&argc, argv);
    if (Initialize (USE_GPU) == GL_NO_ERROR) 
    {
        // create VBO (if using standard GL or CL-GL interop), otherwise create Cl buffer
        createVBO(&vbo);
        
        // initialize the particles
        h_particleData = (float *) malloc (nd * numStars * sizeof(float));
        int pCounter, i;
        for (pCounter = 0; pCounter < numStars; pCounter ++)
            for (i=0; i< nd; i++)
                h_particleData[nd * pCounter + i] = stars[pCounter][i];
        
        // combine creation of memory on the device and copying values over
        cl_membufferData = clCreateBuffer(ComputeContext, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR,
                                          nd * numStars * sizeof(float), h_particleData, &ciErrNum);
        
        int computationWidth = numStars/computationHeight;
        
        // set the args values 
        ciErrNum  = clSetKernelArg(ComputeKernel, 0, sizeof(cl_mem), (void *) &vbo_cl);
        ciErrNum |= clSetKernelArg(ComputeKernel, 1, sizeof(cl_mem), (void *) &cl_membufferData); 
        ciErrNum |= clSetKernelArg(ComputeKernel, 2, sizeof(unsigned int), (void *) &computationWidth);
        ciErrNum |= clSetKernelArg(ComputeKernel, 3, sizeof(unsigned int), (void *) &computationHeight);
        ciErrNum |= clSetKernelArg(ComputeKernel, 4, sizeof(unsigned int), (void *) &numStars);
        //shrCheckErrorEX(ciErrNum, CL_SUCCESS, pCleanup);

        
        
        atexit(Shutdown);
        printf("Starting event loop...\n");

        glutMainLoop();
    }

    return 0;
}

