/* OpenCL version of the andy galaxy code - based on the OpenCL Nvidia simpleGL example
and originally on the CUDA simpleGL example */

// note that the stars really shouldnt collapse to the center etc

#define mass 0.00006
#define stepper (unsigned int) 10

///////////////////////////////////////////////////////////////////////////////
//! version of the galaxy collision code using OpenCL
///////////////////////////////////////////////////////////////////////////////
__kernel void galaxy(__global float4* pos,
                     __global float* pdata,
                     const uint width,
                     const uint height,
                     const uint howManyStars,
                     const uint time)
{
// pos contains vec4 positions for numStars followed by vec4 colours for numStars
// pdata contains 7 floats (x,y,z. vx,vy,vz, mass

    const uint x = get_global_id(0);
    const uint y = get_global_id(1);

	const uint arrayLoc = y*width*7 + x*7;
	const uint posLoc   = y*width+x;
	const uint correspondingLoc = width * height + posLoc;
	
    float4 c = (float4)(1.0, 1.0, 1.0, 1.0);
	float3 affect = (float3)(0.0, 0.0, 0.0);
    float3 newV;
	float3 posit;
	float3 dif;

	//float diag, invDistCube;
	unsigned int i;
    const uint j = time % stepper;
    float d;

if (posLoc < howManyStars)
{	
	
	// if first time through then initialize the position and colour data
    // would be better to initialize this in the host
	if (time < 2)
		{
		pos[posLoc].x = pdata[arrayLoc+0];
		pos[posLoc].y = pdata[arrayLoc+1];
		pos[posLoc].z = pdata[arrayLoc+2];
		pos[posLoc].w = 1.0f; // used to be mass = pdata[arrayLoc+6];

		pos[correspondingLoc] = (float4)(1.0f, 1.0f, 1.0f, 1.0f);
		}

	posit.xyz = pos[posLoc].xyz;
	
	for (i=j; i <width * height; i+=stepper)
		{
		dif.xyz = pos[i].xyz - posit.xyz;
        d = distance(pos[i].xyz, posit.xyz);
 
        d = (d < 1.0f) ? 1.0f : d;

        affect += (float3) 1.0/(d*d*d) * dif; 
 		}
	
	// reading and writing to pdata is not overly cool
    // should really write out a new pdata velocity buffer for the next time step
    // and a newset of positions in the VBO
    
    newV.x = pdata[arrayLoc+3];
    newV.y = pdata[arrayLoc+4];
    newV.z = pdata[arrayLoc+5];


    newV.xyz += (float3) mass * affect.xyz;
    pos[posLoc].xyz += (float3) 0.001f * newV.xyz; 
	    
    c.x = (newV.x < 0.0f) ? 0.4f : 1.0f;
    c.y = (newV.y < 0.0f) ? 0.4f : 1.0f;
    c.z = (newV.z < 0.0f) ? 0.4f : 1.0f;
   
	pos[correspondingLoc] = (float4) c;
    
    pdata[arrayLoc+3] = newV.x;
	pdata[arrayLoc+4] = newV.y;
	pdata[arrayLoc+5] = newV.z;
    
		
	}
}
